"""Test the Structured Output Wrappers."""
import os
import pytest
from pydantic import BaseModel

from sik_llms import (
    create_client,
    system_message,
    user_message,
    StructuredOutputResponse,
)
from tests.conftest import ANTHROPIC_TEST_MODEL, OPENAI_TEST_MODEL
from dotenv import load_dotenv
load_dotenv()


@pytest.mark.asyncio
class TestStructuredOutputs:
    """Test the OpenAI Structured Output Wrapper."""

    @pytest.mark.parametrize('model_name', [
        pytest.param(
            OPENAI_TEST_MODEL,
            id="OpenAI",
        ),
        pytest.param(
            ANTHROPIC_TEST_MODEL,
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    @pytest.mark.parametrize('run_async', [True, False])
    async def test__anthropic__structured_outputs(self, model_name: str, run_async: bool):
        class CalendarEvent(BaseModel):
            name: str
            date: str
            participants: list[str]

        client = create_client(
            model_name=model_name,
            response_format=CalendarEvent,
        )
        messages=[
            system_message("Extract the event information."),
            user_message("Alice and Bob are going to a science fair on Friday."),
        ]
        if run_async:
            response = await client.run_async(messages=messages)
        else:
            response = client(messages=messages)
        assert isinstance(response, StructuredOutputResponse)
        assert isinstance(response.parsed, CalendarEvent)
        assert response.parsed.name
        assert response.parsed.date
        assert response.parsed.participants
        assert 'Alice' in response.parsed.participants
        assert 'Bob' in response.parsed.participants

    @pytest.mark.stochastic(samples=5, threshold=0.5)
    @pytest.mark.parametrize('model_name', [
        pytest.param(
            OPENAI_TEST_MODEL,
            id="OpenAI",
        ),
        pytest.param(
            ANTHROPIC_TEST_MODEL,
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    async def test__anthropic__structured_outputs__nested(self, model_name: str):
        class Address(BaseModel):
            street: str
            street_2: str | None = None
            city: str
            state: str
            zip_code: str

        class Contact(BaseModel):
            first_name: str
            last_name: str
            phone: str | None = None
            email: str | None = None
            address: Address

        client = create_client(
            model_name=model_name,
            response_format=Contact,
        )
        messages=[
            system_message("Extract the information."),
            user_message("Hey my name is Shane Kercheval. I live at 123 Main Street in Anytown, Washington in the USA. The zip code is 12345."),  # noqa: E501
        ]

        response = client(messages=messages)
        assert isinstance(response, StructuredOutputResponse)
        assert isinstance(response.parsed, Contact)
        assert response.parsed.first_name == 'Shane'
        assert response.parsed.last_name == 'Kercheval'
        assert not response.parsed.phone
        assert not response.parsed.email
        assert response.parsed.address.street == '123 Main Street'
        assert not response.parsed.address.street_2
        assert response.parsed.address.city == 'Anytown'
        assert response.parsed.address.state in ('Washington', 'WA')
        assert response.parsed.address.zip_code == '12345'



"""Fixtures for testing the sik_llms module."""
from dataclasses import dataclass
import os
import pytest
from sik_llms import Tool, Parameter
from sik_llms.models_base import RegisteredClients


OPENAI_TEST_MODEL = 'gpt-4o-mini'
OPENAI_TEST_REASONING_MODEL = 'o3-mini'

ANTHROPIC_TEST_MODEL = 'claude-3-5-haiku-latest'
ANTRHOPIC_THINKING_MODEL = 'claude-3-7-sonnet-latest'


@dataclass
class ClientConfig:
    """Configuration for instantiating a client."""

    client_type: RegisteredClients
    model_name: str



@pytest.fixture
def project_root():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))

@pytest.fixture
def test_files_path(project_root: str) -> str:
    return os.path.join(project_root, 'tests', 'test_files')


@pytest.fixture
def simple_weather_tool() -> Tool:
    """Create a simple weather tool with one required parameter."""
    return Tool(
        name="get_weather",
        description="Get the weather for a location.",
        parameters=[
            Parameter(
                name="location",
                param_type=str,
                required=True,
                description="The city and country for weather info.",
            ),
        ],
    )

@pytest.fixture
def complex_weather_tool() -> Tool:
    """Create a weather tool with multiple parameters including an enum."""
    return Tool(
        name="get_detailed_weather",
        description="Get detailed weather information for a location.",
        parameters=[
            Parameter(
                name="location",
                param_type=str,
                required=True,
                description="The city and country",
            ),
            Parameter(
                name="unit",
                param_type=str,
                required=False,
                description="Temperature unit",
                valid_values=["celsius", "fahrenheit"],
            ),
            Parameter(
                name="include_forecast",
                param_type=bool,
                required=False,
                description="Whether to include forecast data",
            ),
        ],
    )

@pytest.fixture
def restaurant_tool() -> Tool:
    """Create a restaurant search tool with multiple parameters."""
    return Tool(
        name="search_restaurants",
        description="Search for restaurants in a location.",
        parameters=[
            Parameter(
                name="location",
                param_type=str,
                required=True,
                description="The city to search in",
            ),
            Parameter(
                name="cuisine",
                param_type=str,
                required=False,
                description="Type of cuisine",
                valid_values=["italian", "chinese", "mexican", "indian"],
            ),
            Parameter(
                name="price_range",
                param_type=str,
                required=False,
                description="Price range",
                valid_values=["$", "$$", "$$$", "$$$$"],
            ),
            Parameter(
                name="open_now",
                param_type=bool,
                required=False,
                description="Filter for currently open restaurants",
            ),
        ],
    )



"""Tests for the Tools Wrappers."""
import asyncio
import os
import pytest
from dotenv import load_dotenv
from enum import Enum
from typing import List, Dict, Literal, Optional, Any, Union  # noqa: UP035
from pydantic import BaseModel, Field
from sik_llms import (
    Client,
    user_message,
    Tool,
    Parameter,
    ToolPredictionResponse,
    ToolPrediction,
    RegisteredClients,
    ToolChoice,
)
from sik_llms.models_base import pydantic_model_to_parameters
from tests.conftest import ANTHROPIC_TEST_MODEL, OPENAI_TEST_MODEL, ClientConfig

load_dotenv()


class TestPydanticModelToParameters:
    """Tests for pydantic_model_to_parameters function."""

    def test_simple_model(self):
        """Test conversion of a simple Pydantic model."""
        class SimpleModel(BaseModel):
            name: str
            age: int
            active: bool
            a_float: float

        params = pydantic_model_to_parameters(SimpleModel)

        # Check number of parameters
        assert len(params) == 4

        # Check parameter properties
        name_param = next(p for p in params if p.name == 'name')
        assert name_param.param_type is str
        assert name_param.required is True

        age_param = next(p for p in params if p.name == 'age')
        assert age_param.param_type is int
        assert age_param.required is True

        active_param = next(p for p in params if p.name == 'active')
        assert active_param.param_type is bool
        assert active_param.required is True

    def test_primitive_types_to_json_schema(self):
        """Test that Python primitive types are converted to correct JSON Schema types."""
        # Test all primitive types
        class AllTypes(BaseModel):
            a_str: str
            an_int: int
            a_float: float
            a_bool: bool
            a_list: list[str]
            a_dict: dict[str, int]

        params = pydantic_model_to_parameters(AllTypes)
        tool = Tool(
            name="test_types",
            parameters=params,
            description="Test type conversions",
            func=lambda **kwargs: str(kwargs),
        )

        # Get the OpenAI format
        openai_format = tool.to_openai()
        properties = openai_format["function"]["parameters"]["properties"]

        # Check each type conversion
        assert properties["a_str"]["type"] == "string"
        assert properties["an_int"]["type"] == "integer"
        assert properties["a_float"]["type"] == "number"
        assert properties["a_bool"]["type"] == "boolean"
        assert properties["a_list"]["type"] == "array"
        assert properties["a_list"]["items"]["type"] == "string"
        assert properties["a_dict"]["type"] == "object"
        assert properties["a_dict"]["additionalProperties"]["type"] == "integer"

        # Do the same check for Anthropic format
        anthropic_format = tool.to_anthropic()
        properties = anthropic_format["input_schema"]["properties"]

        assert properties["a_str"]["type"] == "string"
        assert properties["an_int"]["type"] == "integer"
        assert properties["a_float"]["type"] == "number"
        assert properties["a_bool"]["type"] == "boolean"
        assert properties["a_list"]["type"] == "array"
        assert properties["a_list"]["items"]["type"] == "string"
        assert properties["a_dict"]["type"] == "object"
        assert properties["a_dict"]["additionalProperties"]["type"] == "integer"

    def test_model_with_defaults(self):
        """Test conversion of a model with default values."""
        class ModelWithDefaults(BaseModel):
            name: str
            age: int = 30
            active: bool = True

        params = pydantic_model_to_parameters(ModelWithDefaults)

        # Check required status
        name_param = next(p for p in params if p.name == 'name')
        assert name_param.required is True

        age_param = next(p for p in params if p.name == 'age')
        assert age_param.required is False, "Fields with defaults should not be marked as required"

        active_param = next(p for p in params if p.name == 'active')
        assert active_param.required is False, "Fields with defaults should not be marked as required"  # noqa: E501

    def test_nested_model_with_defaults(self):
        """Test conversion of nested models with default values."""
        class Address(BaseModel):
            street: str
            city: str
            country: str = 'USA'  # Default value

        class Person(BaseModel):
            name: str
            phone: str = Field(default='<not provided>')
            address: Address

        params = pydantic_model_to_parameters(Person)

        # Check required status
        name_param = next(p for p in params if p.name == 'name')
        assert name_param.required is True

        phone_param = next(p for p in params if p.name == 'phone')
        assert phone_param.required is False

        address_param = next(p for p in params if p.name == 'address')
        assert address_param.required is True

        # Convert to OpenAI format and check for no defaults
        tool = Tool(
            name='test_person',
            parameters=params,
            description='Test nested defaults',
        )

        openai_format = tool.to_openai()
        schema = openai_format['function']['parameters']['properties']
        print(schema)
        # Verify no default values in schema
        assert 'default' not in schema['phone']
        assert 'street' in schema['address']['properties']
        assert 'city' in schema['address']['properties']
        assert 'country' in schema['address']['properties']
        assert 'default' not in schema['address']['properties']['country']

    def test_complex_nested_objects_with_defaults(self):
        """Test handling of complex nested objects with defaults at multiple levels."""
        class Configuration(BaseModel):
            debug: bool = False
            verbosity: int = 1

        class Database(BaseModel):
            url: str
            timeout: int = 30
            config: Configuration = Field(default_factory=Configuration)

        class AppSettings(BaseModel):
            app_name: str
            db: Database
            temp_dir: str = "/tmp"
            backup_dbs: list[Database] = Field(default_factory=list)

        params = pydantic_model_to_parameters(AppSettings)

        # Check required fields
        app_name_param = next(p for p in params if p.name == 'app_name')
        assert app_name_param.required is True

        db_param = next(p for p in params if p.name == 'db')
        assert db_param.required is True

        temp_dir_param = next(p for p in params if p.name == 'temp_dir')
        assert temp_dir_param.required is False

        backup_dbs_param = next(p for p in params if p.name == 'backup_dbs')
        assert backup_dbs_param.required is False

        # Convert to OpenAI format
        tool = Tool(
            name="app_settings",
            parameters=params,
            description="Test app settings",
        )

        openai_format = tool.to_openai()

        # Verify no defaults exist anywhere in the schema
        def check_no_defaults(obj) -> None:  # noqa: ANN001
            if isinstance(obj, dict):
                assert "default" not in obj, f"Found 'default' in: {obj}"
                for key, value in obj.items():
                    if isinstance(value, dict | list):
                        check_no_defaults(value)
            elif isinstance(obj, list):
                for item in obj:
                    if isinstance(item, dict | list):
                        check_no_defaults(item)

        check_no_defaults(openai_format)

    def test_model_with_optional_fields(self):
        """Test conversion of a model with Optional fields."""
        class ModelWithOptional(BaseModel):
            name: str
            age: Optional[int]  # Using Optional syntax  # noqa: UP007
            email: str | None   # Using pipe syntax

        ModelWithOptional(name='shane', email=None, age=None)

        params = pydantic_model_to_parameters(ModelWithOptional)

        # Check required status
        name_param = next(p for p in params if p.name == 'name')
        assert name_param.required is True

        age_param = next(p for p in params if p.name == 'age')
        assert age_param.param_type is int
        assert age_param.required is True

        email_param = next(p for p in params if p.name == 'email')
        assert email_param.param_type is str
        assert email_param.required is True

    def test_model_with_descriptions(self):
        """Test conversion of a model with field descriptions."""
        class ModelWithDescriptions(BaseModel):
            name: str = Field(description="The person's full name")
            age: int = Field(description='Age in years')

        params = pydantic_model_to_parameters(ModelWithDescriptions)

        name_param = next(p for p in params if p.name == 'name')
        assert name_param.description == "The person's full name"

        age_param = next(p for p in params if p.name == 'age')
        assert age_param.description == 'Age in years'

    def test_model_with_complex_types(self):
        """Test conversion of a model with complex field types."""
        class ModelWithComplexTypes(BaseModel):
            tags: List[str]  # noqa: UP006
            counts: Dict[str, int]  # noqa: UP006

        params = pydantic_model_to_parameters(ModelWithComplexTypes)

        tags_param = next(p for p in params if p.name == 'tags')
        assert tags_param.param_type == List[str]  # Direct Python type comparison  # noqa: UP006

        counts_param = next(p for p in params if p.name == 'counts')
        assert counts_param.param_type == Dict[str, int]  # Direct Python type comparison  # noqa: E501, UP006

    def test_model_with_enum(self):
        """Test conversion of a model with Enum fields."""
        class Color(Enum):
            RED = 'red'
            GREEN = 'green'
            BLUE = 'blue'

        class ModelWithEnum(BaseModel):
            color: Color

        params = pydantic_model_to_parameters(ModelWithEnum)

        color_param = params[0]
        assert color_param.param_type is str
        assert color_param.valid_values == ['red', 'green', 'blue']

    # Better test for nested models
    def test_nested_models(self):
        """Test conversion of nested Pydantic models."""
        class Address(BaseModel):
            street: str
            city: str
            zip_code: str

        class Person(BaseModel):
            name: str
            address: Address = Field(description="The person's address")

        person_params = pydantic_model_to_parameters(Person)
        assert person_params

        assert person_params[0].name == 'name'
        assert person_params[0].param_type is str
        assert person_params[0].required is True
        assert person_params[0].description is None
        assert person_params[0].valid_values is None

        assert person_params[1].name == 'address'
        assert person_params[1].param_type is Address
        assert person_params[1].required is True
        assert person_params[1].description is not None
        assert "The person's address" in person_params[1].description
        assert person_params[1].valid_values is None

    # Better test for list of models
    def test_model_with_list_of_models(self):
        """Test conversion of a model with a list of nested models."""
        class Item(BaseModel):
            name: str
            price: float

        class Order(BaseModel):
            order_id: str
            items: List[Item]  # noqa: UP006

        params = pydantic_model_to_parameters(Order)

        items_param = next(p for p in params if p.name == 'items')
        assert items_param.param_type == List[Item]  # noqa: UP006
        assert items_param.description is None

    def test_model_with_unsupported_types(self):
        """Test handling of models with unsupported types."""
        class ModelWithUnsupportedTypes(BaseModel):
            name: str
            anything: Any  # Unsupported type

        with pytest.raises(ValueError):  # noqa: PT011
            pydantic_model_to_parameters(ModelWithUnsupportedTypes)

    def test_model_with_optional_complex_types(self):
        """Test conversion of a model with optional complex types."""
        class ModelWithOptionalComplex(BaseModel):
            tags: list[str] | None = None
            metadata: dict[str, str] | None = None

        params = pydantic_model_to_parameters(ModelWithOptionalComplex)

        tags_param = next(p for p in params if p.name == 'tags')
        assert tags_param.param_type == list[str]
        assert tags_param.required is False

        metadata_param = next(p for p in params if p.name == 'metadata')
        assert metadata_param.param_type == dict[str, str]
        assert metadata_param.required is False

    # Add new tests for our updated features
    def test_model_with_union_types(self):
        """Test conversion of a model with union types."""
        class ModelWithUnion(BaseModel):
            id: int | str
            data: Dict[str, int] | List[str]  # noqa: UP006

        params = pydantic_model_to_parameters(ModelWithUnion)

        id_param = next(p for p in params if p.name == 'id')
        assert id_param.param_type is str  # Base type is string (placeholder)
        assert id_param.any_of == [int, str]  # Should contain both union types

        data_param = next(p for p in params if p.name == 'data')
        assert data_param.param_type is str  # Base type is string (placeholder)
        assert len(data_param.any_of) == 2
        assert Dict[str, int] in data_param.any_of  # noqa: UP006
        assert List[str] in data_param.any_of  # noqa: UP006

    def test_model_with_literal_types(self):
        """Test conversion of a model with Literal types."""
        class ModelWithLiteral(BaseModel):
            status: Literal["pending", "approved", "rejected"]
            priority: Literal[1, 2, 3]

        params = pydantic_model_to_parameters(ModelWithLiteral)

        status_param = next(p for p in params if p.name == 'status')
        assert status_param.param_type is str
        assert set(status_param.valid_values) == {"pending", "approved", "rejected"}

        priority_param = next(p for p in params if p.name == 'priority')
        assert priority_param.param_type is int
        assert set(priority_param.valid_values) == {1, 2, 3}

    def test_model_with_mixed_literal_types(self):
        """Test handling of Literal types with mixed value types."""
        class MixedLiteralModel(BaseModel):
            mixed: Literal["success", 404, True]  # String, int, and bool

        params = pydantic_model_to_parameters(MixedLiteralModel)
        mixed_param = params[0]

        # Since mixed types are converted to strings
        assert mixed_param.param_type is str
        assert set(mixed_param.valid_values) == {"success", "404", "True"}

    def test_model_with_complex_nested_unions(self):
        """Test handling of complex nested union types."""
        class ComplexUnionModel(BaseModel):
            data: Union[str, List[int], Dict[str, bool]]  # noqa: UP006, UP007
            data2: str | list[int] | dict[str, bool]

        params = pydantic_model_to_parameters(ComplexUnionModel)
        data_param = params[0]

        assert data_param.param_type is str  # Base placeholder type
        assert len(data_param.any_of) == 3
        assert str in data_param.any_of
        assert List[int] in data_param.any_of  # noqa: UP006
        assert Dict[str, bool] in data_param.any_of  # noqa: UP006

    def test_parameter_with_param_type_validation(self):
        """Test Parameter class validation for param_type."""
        # Valid param_type values
        Parameter(name="test1", param_type=str, required=True)
        Parameter(name="test2", param_type=int, required=True)
        Parameter(name="test3", param_type=list[str], required=True)
        Parameter(name="test4", param_type=dict[str, int], required=True)

        class TestModel(BaseModel):
            x: int

        Parameter(name="test5", param_type=TestModel, required=True)

        # Invalid param_type values
        with pytest.raises(ValueError):  # noqa: PT011
            Parameter(name="invalid1", param_type=Any, required=True)

        with pytest.raises(ValueError):  # noqa: PT011
            Parameter(name="invalid2", param_type="string", required=True)  # `str` instead of type

    def test_parameter_serialization(self):
        """Test Parameter serialization and deserialization."""
        param = Parameter(
            name="test",
            param_type=str,
            required=True,
            description="Test parameter",
            valid_values=["a", "b", "c"],
        )

        # Serialize
        data = param.model_dump()

        # Verify serialized data
        assert data["name"] == "test"
        assert "param_type" in data  # Now stored as string representation
        assert data["required"] is True
        assert data["description"] == "Test parameter"
        assert data["valid_values"] == ["a", "b", "c"]

        # Deserialize (simplified test)
        deserialized = Parameter.model_validate(data)

        # Basic verification of deserialized object
        assert deserialized.name == param.name
        assert deserialized.required == param.required
        assert deserialized.description == param.description
        assert deserialized.valid_values == param.valid_values
        # Note: param_type may not match exactly due to serialization/deserialization

    def test_parameter_to_openai_conversion(self):
        """Test conversion of Parameter objects to OpenAI format."""
        # Test string parameter with valid_values
        param = Parameter(
            name="color",
            param_type=str,
            required=True,
            description="Color selection",
            valid_values=["red", "green", "blue"],
        )

        tool = Tool(
            name="set_color",
            parameters=[param],
            description="Set a color",
            func=lambda color: f"Color set to {color}",
        )

        openai_format = tool.to_openai()
        assert openai_format["function"]["parameters"]["properties"]["color"]["type"] == "string"
        assert openai_format["function"]["parameters"]["properties"]["color"]["enum"] == ["red", "green", "blue"]  # noqa: E501
        assert openai_format["function"]["parameters"]["properties"]["color"]["description"] == "Color selection"  # noqa: E501

    def test_parameter_to_openai_with_any_of(self):
        """Test conversion of Parameter with any_of to OpenAI format."""
        param = Parameter(
            name="id",
            param_type=str,  # Base type
            required=True,
            description="ID value",
            any_of=[str, int],
        )

        tool = Tool(
            name="get_item",
            parameters=[param],
            description="Get an item",
            func=lambda id: f"Item {id}",  # noqa: A006
        )

        openai_format = tool.to_openai()
        assert "anyOf" in openai_format["function"]["parameters"]["properties"]["id"]
        assert len(openai_format["function"]["parameters"]["properties"]["id"]["anyOf"]) == 2
        assert openai_format["function"]["parameters"]["properties"]["id"]["description"] == "ID value"  # noqa: E501
        assert openai_format['function']["description"] == "Get an item"

    def test_openai_schema_has_no_defaults(self):
        """Test that the OpenAI schema doesn't contain any default values."""
        class ConfigModel(BaseModel):
            name: str
            timeout: int = 30
            retries: int = 3
            settings: dict = Field(default_factory=dict)

        params = pydantic_model_to_parameters(ConfigModel)
        tool = Tool(
            name="test_config",
            parameters=params,
            description="Test config",
        )

        # Convert to OpenAI format
        openai_format = tool.to_openai()
        properties = openai_format["function"]["parameters"]["properties"]

        # Check that no properties have default values
        for prop_name, prop_schema in properties.items():
            assert "default" not in prop_schema, f"Property {prop_name} should not have a default value"  # noqa: E501

        # Check required fields
        required = openai_format["function"]["parameters"].get("required", [])
        assert "name" in required
        assert "timeout" not in required
        assert "retries" not in required
        assert "settings" not in required

    def test_parameter_to_anthropic_conversion(self):
        """Test conversion of Parameter objects to Anthropic format."""
        param = Parameter(
            name="color",
            param_type=str,
            required=True,
            description="Color selection",
            valid_values=["red", "green", "blue"],
        )

        tool = Tool(
            name="set_color",
            parameters=[param],
            description="Set a color",
            func=lambda color: f"Color set to {color}",
        )

        anthropic_format = tool.to_anthropic()
        assert anthropic_format["name"] == "set_color"
        assert anthropic_format["description"] == "Set a color"
        assert anthropic_format["input_schema"]["properties"]["color"]["type"] == "string"
        assert anthropic_format["input_schema"]["properties"]["color"]["enum"] == ["red", "green", "blue"]  # noqa: E501
        assert anthropic_format["input_schema"]["properties"]["color"]["description"] == "Color selection"  # noqa: E501

    def test_parameter_to_anthropic_with_any_of(self):
        """Test conversion of Parameter with any_of to Anthropic format."""
        param = Parameter(
            name="id",
            param_type=str,  # Base type
            required=True,
            description="ID value",
            any_of=[str, int],
        )

        tool = Tool(
            name="get_item",
            parameters=[param],
            description="Get an item",
            func=lambda id: f"Item {id}",  # noqa: A006
        )

        anthropic_format = tool.to_anthropic()
        assert "anyOf" in anthropic_format["input_schema"]["properties"]["id"]
        assert len(anthropic_format["input_schema"]["properties"]["id"]["anyOf"]) == 2
        assert anthropic_format["input_schema"]["properties"]["id"]["description"] == "ID value"

    def test_anthropic_schema_has_no_defaults(self):
        """Test that the Anthropic schema doesn't contain any default values."""
        class ConfigModel(BaseModel):
            name: str
            timeout: int = 30
            retries: int = 3
            settings: dict = Field(default_factory=dict)

        params = pydantic_model_to_parameters(ConfigModel)
        tool = Tool(
            name="test_config",
            parameters=params,
            description="Test config",
        )

        # Convert to Anthropic format
        anthropic_format = tool.to_anthropic()
        properties = anthropic_format["input_schema"]["properties"]

        # Check that no properties have default values
        for prop_name, prop_schema in properties.items():
            assert "default" not in prop_schema, f"Property {prop_name} should not have a default value"  # noqa: E501

        # Check required fields
        required = anthropic_format["input_schema"].get("required", [])
        assert "name" in required
        assert "timeout" not in required
        assert "retries" not in required
        assert "settings" not in required


@pytest.mark.asyncio
class TestTools:
    """Test the Tools Wrapper."""

    # @pytest.mark.stochastic(samples=5, threshold=0.5)
    @pytest.mark.parametrize('client_config', [
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.OPENAI_TOOLS,
                model_name=OPENAI_TEST_MODEL,
            ),
            id="OpenAI",
        ),
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.ANTHROPIC_TOOLS,
                model_name=ANTHROPIC_TEST_MODEL,
            ),
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    @pytest.mark.parametrize('is_async', [
        pytest.param(True, id="Async"),
        pytest.param(False, id="Sync"),
    ])
    async def test__single_tool_single_parameter__instantiate(
            self,
            simple_weather_tool: Tool,
            is_async: bool,
            client_config: ClientConfig,
        ):
        """Test calling a simple tool with one required parameter."""
        client = Client.instantiate(
            client_type=client_config.client_type,
            model_name=client_config.model_name,
            tools=[simple_weather_tool],
        )
        if is_async:
            response = await client.run_async(
                messages=[
                    user_message("What's the weather like in Paris?"),
                ],
            )
        else:
            response = client(
                messages=[
                    user_message("What's the weather like in Paris?"),
                ],
            )
        assert isinstance(response, ToolPredictionResponse)
        assert isinstance(response.tool_prediction, ToolPrediction)
        assert response.tool_prediction.name == "get_weather"
        assert "location" in response.tool_prediction.arguments
        assert "Paris" in response.tool_prediction.arguments["location"]
        assert response.input_tokens > 0
        assert response.output_tokens > 0
        assert response.input_cost > 0
        assert response.output_cost > 0

    @pytest.mark.parametrize('client_config', [
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.OPENAI_TOOLS,
                model_name=OPENAI_TEST_MODEL,
            ),
            id="OpenAI",
        ),
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.ANTHROPIC_TOOLS,
                model_name=ANTHROPIC_TEST_MODEL,
            ),
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    async def test__no_tool_call(
            self,
            simple_weather_tool: Tool,
            client_config: ClientConfig,
        ):
        """Test calling a simple tool when no tool is applicable."""
        client = Client.instantiate(
            client_type=client_config.client_type,
            model_name=client_config.model_name,
            tools=[simple_weather_tool],
            tool_choice=ToolChoice.AUTO,
        )
        response = await client.run_async(
            messages=[
                user_message("What's the stock price of Apple?"),
            ],
        )
        assert isinstance(response, ToolPredictionResponse)
        assert response.tool_prediction is None
        assert response.message
        assert response.input_tokens > 0
        assert response.output_tokens > 0
        assert response.input_cost > 0
        assert response.output_cost > 0

    @pytest.mark.parametrize('client_config', [
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.OPENAI_TOOLS,
                model_name=OPENAI_TEST_MODEL,
            ),
            id="OpenAI",
        ),
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.ANTHROPIC_TOOLS,
                model_name=ANTHROPIC_TEST_MODEL,
            ),
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    @pytest.mark.parametrize('is_async', [
        pytest.param(True, id="Async"),
        pytest.param(False, id="Sync"),
    ])
    async def test__single_tool_multiple_parameters(
            self,
            complex_weather_tool: Tool,
            is_async: bool,
            client_config: ClientConfig,
        ):
        """Test calling a tool with multiple parameters including optional ones."""
        client = Client.instantiate(
            client_type=client_config.client_type,
            model_name=client_config.model_name,
            tools=[complex_weather_tool],
        )
        if is_async:
            response = await client.run_async(
                messages=[
                    user_message("What's the weather like in Tokyo in Celsius with forecast?"),
                ],
            )
        else:
            response = client(
                messages=[
                    user_message("What's the weather like in Tokyo in Celsius with forecast?"),
                ],
            )
        assert response.tool_prediction.name == "get_detailed_weather"
        args = response.tool_prediction.arguments
        assert "Tokyo" in args["location"]
        assert args.get("unit") in ["celsius", "fahrenheit"]
        assert args.get("include_forecast") is not None

    @pytest.mark.parametrize('client_config', [
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.OPENAI_TOOLS,
                model_name=OPENAI_TEST_MODEL,
            ),
            id="OpenAI",
        ),
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.ANTHROPIC_TOOLS,
                model_name=ANTHROPIC_TEST_MODEL,
            ),
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    @pytest.mark.parametrize('is_async', [
        pytest.param(True, id="Async"),
        pytest.param(False, id="Sync"),
    ])
    async def test__multiple_tools(
            self,
            simple_weather_tool: Tool,
            restaurant_tool: Tool,
            is_async: bool,
            client_config: ClientConfig,
        ):
        """Test providing multiple tools to the model."""
        client = Client.instantiate(
            client_type=client_config.client_type,
            model_name=client_config.model_name,
            tools=[simple_weather_tool, restaurant_tool],
        )
        # Weather query
        if is_async:
            weather_response = await client.run_async(
                messages=[
                    user_message("What's the weather like in London?"),
                ],
            )
        else:
            weather_response = client(
                messages=[
                    user_message("What's the weather like in London?"),
                ],
            )
        assert weather_response.tool_prediction.name == "get_weather"
        assert "London" in weather_response.tool_prediction.arguments["location"]
        # Restaurant query
        if is_async:
            restaurant_response = await client.run_async(
                messages=[
                    user_message("Find me an expensive Italian restaurant in New York"),
                ],
            )
        else:
            restaurant_response = client(
                messages=[
                    user_message("Find me an expensive Italian restaurant in New York"),
                ],
            )
        assert restaurant_response.tool_prediction.name == "search_restaurants"
        args = restaurant_response.tool_prediction.arguments
        assert "New York" in args["location"]
        assert args.get("cuisine") == "italian"
        assert args.get("price_range") in ["$$", "$$$", "$$$$"]

    @pytest.mark.parametrize('client_config', [
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.OPENAI_TOOLS,
                model_name=OPENAI_TEST_MODEL,
            ),
            id="OpenAI",
        ),
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.ANTHROPIC_TOOLS,
                model_name=ANTHROPIC_TEST_MODEL,
            ),
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    @pytest.mark.parametrize('is_async', [
        pytest.param(True, id="Async"),
        pytest.param(False, id="Sync"),
    ])
    async def test__enum_parameters(
            self,
            restaurant_tool: Tool,
            is_async: bool,
            client_config: ClientConfig,
        ):
        """Test handling of enum parameters."""
        client = Client.instantiate(
            client_type=client_config.client_type,
            model_name=client_config.model_name,
            tools=[restaurant_tool],
        )
        test_cases = [
            (
                "Find me a cheap Chinese restaurant in Boston",
                {"cuisine": "chinese", "price_range": "$"},
            ),
            (
                "I want Mexican food in Chicago at a moderate price",
                {"cuisine": "mexican"},
            ),
            (
                "Find an Indian restaurant in Seattle",
                {"cuisine": "indian"},
            ),
        ]

        for prompt, expected_args in test_cases:
            if is_async:
                response = await client.run_async(messages=[user_message(prompt)])
            else:
                response = client(messages=[user_message(prompt)])
            args = response.tool_prediction.arguments
            for key, value in expected_args.items():
                assert args.get(key) == value

    @pytest.mark.parametrize('client_config', [
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.OPENAI_TOOLS,
                model_name=OPENAI_TEST_MODEL,
            ),
            id="OpenAI",
        ),
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.ANTHROPIC_TOOLS,
                model_name=ANTHROPIC_TEST_MODEL,
            ),
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    async def test__concurrent_tool_calls(self, simple_weather_tool: Tool, client_config: ClientConfig):  # noqa: E501
        """Test multiple concurrent tool calls."""
        client = Client.instantiate(
            client_type=client_config.client_type,
            model_name=client_config.model_name,
            tools=[simple_weather_tool],
        )

        cities = ["Paris", "London", "Tokyo", "New York", "Sydney"]
        messages = [
            [{"role": "user", "content": f"What's the weather like in {city}?"}]
            for city in cities
        ]

        responses = await asyncio.gather(*(
            client.run_async(messages=msg) for msg in messages
        ))

        for i, response in enumerate(responses):
            assert response.tool_prediction.name == "get_weather"
            assert cities[i] in response.tool_prediction.arguments["location"]
            assert response.input_tokens > 0
            assert response.output_tokens > 0

    @pytest.mark.parametrize('client_config', [
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.OPENAI_TOOLS,
                model_name=OPENAI_TEST_MODEL,
            ),
            id="OpenAI",
        ),
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.ANTHROPIC_TOOLS,
                model_name=ANTHROPIC_TEST_MODEL,
            ),
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    @pytest.mark.parametrize(('prompt', 'expected_status'), [
        pytest.param("Update order #12345 to shipped status", 'shipped', id='shipped'),
        pytest.param("Cancel my order ABC-789", 'cancelled', id='cancelled'),
        pytest.param("Mark order XYZ-456 as delivered", 'delivered', id='delivered'),
        pytest.param("Set order 55555 to processing", 'processing', id='processing'),
    ])
    async def test_tool_with_valid_values(
        self,
        prompt: str,
        expected_status: str,
        client_config: ClientConfig,
    ):
        """Test a tool with parameters that have valid_values constraints."""
        # Define a tool with valid_values (enum-like) constraints
        status_tool = Tool(
            name="update_status",
            description="Update the status of an order",
            parameters=[
                Parameter(
                    name="order_id",
                    param_type=str,
                    required=True,
                    description="The ID of the order to update",
                ),
                Parameter(
                    name="status",
                    param_type=str,
                    required=True,
                    description="The new status for the order",
                    valid_values=["pending", "processing", "shipped", "delivered", "cancelled"],
                ),
            ],
            func=lambda order_id, status: f"Order {order_id} updated to {status}",
        )
        client = Client.instantiate(
            client_type=client_config.client_type,
            model_name=client_config.model_name,
            tools=[status_tool],
        )
        response = await client.run_async(
            messages=[user_message(prompt)],
        )
        assert response.tool_prediction.name == "update_status"
        args = response.tool_prediction.arguments
        assert "order_id" in args
        assert "status" in args
        assert args["status"] == expected_status

    @pytest.mark.parametrize('client_config', [
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.OPENAI_TOOLS,
                model_name=OPENAI_TEST_MODEL,
            ),
            id="OpenAI",
        ),
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.ANTHROPIC_TOOLS,
                model_name=ANTHROPIC_TEST_MODEL,
            ),
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    @pytest.mark.parametrize('test_case', [
        pytest.param(
            {
                "prompt": "Look up product with ID 78901",
                "expected_id": 78901,
                "expected_id_type": int,
            },
            id="integer",
        ),
        pytest.param(
            {
                "prompt": "Get details for SKU abc-123",
                "expected_id": "abc-123",
                "expected_id_type": str,
            },
            id="string",
        ),
    ])
    async def test_tool_with_any_of_parameters(
        self,
        test_case: dict,
        client_config: ClientConfig,
        ):
        """Test a tool with parameters that use any_of for union types."""
        # Define a tool with any_of (union type) constraints
        prompt = test_case["prompt"]
        expected_id = test_case["expected_id"]
        expected_id_type = test_case["expected_id_type"]
        search_tool = Tool(
            name="search_item",
            description="Search for an item by ID or name",
            parameters=[
                Parameter(
                    name="identifier",
                    param_type=str,  # Base type
                    required=True,
                    description="The item identifier (can be numeric ID or name string)",
                    any_of=[str, int],  # Can be string or integer
                ),
                Parameter(
                    name="category",
                    param_type=str,
                    required=False,
                    description="Item category to filter by",
                ),
            ],
            func=lambda identifier, category=None: f"Searching for {identifier} in {category or 'all categories'}",  # noqa: E501
        )
        client = Client.instantiate(
            client_type=client_config.client_type,
            model_name=client_config.model_name,
            tools=[search_tool],
        )
        response = await client.run_async(
            messages=[user_message(prompt)],
        )
        assert response.tool_prediction.name == "search_item"
        args = response.tool_prediction.arguments
        assert "identifier" in args

        # Check that the identifier is of the expected type
        assert isinstance(args["identifier"], expected_id_type)
        # Check that it contains the expected value
        if expected_id_type is int:
            assert args["identifier"] == expected_id
        else:
            assert expected_id in args["identifier"]

    @pytest.mark.parametrize('client_config', [
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.OPENAI_TOOLS,
                model_name=OPENAI_TEST_MODEL,
            ),
            id="OpenAI",
        ),
        pytest.param(
            ClientConfig(
                client_type=RegisteredClients.ANTHROPIC_TOOLS,
                model_name=ANTHROPIC_TEST_MODEL,
            ),
            id="Anthropic",
            marks=pytest.mark.skipif(
                os.getenv('ANTHROPIC_API_KEY') is None,
                reason="ANTHROPIC_API_KEY is not set",
            ),
        ),
    ])
    @pytest.mark.parametrize('test_case', [
        pytest.param(
            {
                "prompt": "Calculate shipping cost for products 1001, 1002, and 1003 to 123 Main St, Boston, MA 02108 using express shipping",  # noqa: E501
                "expected_products": [1001, 1002, 1003],
                "expected_city": "Boston",
                "expected_method": "express",
            },
            id="multiple_products_express",
        ),
        pytest.param(
            {
                "prompt": "What's the shipping cost for item 5000 to 456 Park Ave, New York, NY 10022 with standard delivery?",  # noqa: E501
                "expected_products": [5000],
                "expected_city": "New York",
                "expected_method": "standard",
            },
            id="single_product_standard",
        ),
        pytest.param(
            {
                "prompt": "I need overnight shipping for products 7777 and 8888 to 789 Ocean Blvd, Miami, FL 33139",  # noqa: E501
                "expected_products": [7777, 8888],
                "expected_city": "Miami",
                "expected_method": "overnight",
            },
            id="two_products_overnight",
        ),
    ])
    async def test_tool_with_complex_nested_parameters(
        self,
        test_case: dict,
        client_config: ClientConfig,
    ):
        """Test a tool with complex nested parameter structures."""
        # Extract test case values
        prompt = test_case["prompt"]
        expected_products = test_case["expected_products"]
        expected_city = test_case["expected_city"]
        expected_method = test_case["expected_method"]
        # Define a Pydantic model for address
        class Address(BaseModel):
            street: str
            city: str
            zip_code: str
            country: str = "USA"

        # Define a tool that takes a complex parameter
        shipping_tool = Tool(
            name="calculate_shipping",
            description="Calculate shipping cost for an order",
            parameters=[
                Parameter(
                    name="product_ids",
                    param_type=list[int],
                    required=True,
                    description="List of product IDs in the order",
                ),
                Parameter(
                    name="shipping_address",
                    param_type=Address,
                    required=True,
                    description="The delivery address",
                ),
                Parameter(
                    name="shipping_method",
                    param_type=str,
                    required=True,
                    description="Shipping method to use",
                    valid_values=["standard", "express", "overnight"],
                ),
            ],
            func=lambda product_ids, shipping_address, shipping_method: f"Shipping cost for {len(product_ids)} items to {shipping_address.city}: ${len(product_ids) * 5} via {shipping_method}",  # noqa: E501
        )

        client = Client.instantiate(
            client_type=client_config.client_type,
            model_name=client_config.model_name,
            tools=[shipping_tool],
        )
        response = await client.run_async(
            messages=[user_message(prompt)],
        )

        assert response.tool_prediction.name == "calculate_shipping"
        args = response.tool_prediction.arguments

        # Check product_ids
        assert "product_ids" in args
        assert isinstance(args["product_ids"], list)
        assert all(isinstance(_id, int) for _id in args["product_ids"])
        # Check that all expected products are included
        for product_id in expected_products:
            assert product_id in args["product_ids"]

        # Check shipping_address
        assert "shipping_address" in args
        address = args["shipping_address"]
        assert "street" in address
        assert "city" in address
        assert expected_city in address["city"]
        assert "zip_code" in address

        # Check shipping_method
        assert "shipping_method" in args
        assert args["shipping_method"] == expected_method



"""Helper functions for OpenAI API."""
from copy import deepcopy
from functools import cache
import json
import os
import time
from collections.abc import AsyncGenerator
from openai import AsyncOpenAI
from pydantic import BaseModel
import tiktoken
from tiktoken import Encoding
from sik_llms.models_base import (
    Tool,
    ToolPredictionResponse,
    ToolPrediction,
    Client,
    TextChunkEvent,
    TextResponse,
    ReasoningEffort,
    RegisteredClients,
    StructuredOutputResponse,
    ToolChoice,
)
from dotenv import load_dotenv
load_dotenv()


CHAT_MODEL_COST_PER_TOKEN = {
    # minor versions
    'gpt-4o-2024-05-13': {
        'input': 5.00 / 1_000_000, 'output': 15.00 / 1_000_000,
        'cached': 1.25 / 1_000_000,
    },
    'gpt-4o-2024-08-06': {
        'input': 2.50 / 1_000_000, 'output': 10.00 / 1_000_000,
        'cached': 1.25 / 1_000_000,
    },
    'gpt-4o-2024-11-20': {
        'input': 2.50 / 1_000_000, 'output': 10.00 / 1_000_000,
        'cached': 1.25 / 1_000_000,
    },
    'gpt-4o-mini-2024-07-18':  {
        'input': 0.15 / 1_000_000, 'output': 0.60 / 1_000_000,
        'cached': 0.075 / 1_000_000,
    },
    'o1-2024-12-17': {
        'input': 15.00 / 1_000_000, 'output': 60.00 / 1_000_000,
        'cached': 7.50 / 1_000_000,
    },
    'o3-mini-2025-01-31': {
        'input': 1.10 / 1_000_000, 'output': 4.40 / 1_000_000,
        'cached': 0.55 / 1_000_000,
    },
    # LEGACY MODELS
    'gpt-4-turbo': {'input': 10.00 / 1_000_000, 'output': 30.00 / 1_000_000},
    'gpt-4-turbo-2024-04-09': {'input': 10.00 / 1_000_000, 'output': 30.00 / 1_000_000},
    'gpt-4-0125-preview': {'input': 0.01 / 1_000, 'output': 0.03 / 1_000},
    'gpt-3.5-turbo': {'input': 0.50 / 1_000_000, 'output': 1.50 / 1_000_000},
    'gpt-3.5-turbo-0125': {'input': 0.50 / 1_000_000, 'output': 1.50 / 1_000_000},
    'gpt-4-0613': {'input': 0.03 / 1_000, 'output': 0.06 / 1_000},
}
CHAT_MODEL_COST_PER_TOKEN_PRIMARY = {
    'gpt-4o-mini': CHAT_MODEL_COST_PER_TOKEN['gpt-4o-mini-2024-07-18'],
    'gpt-4o': CHAT_MODEL_COST_PER_TOKEN['gpt-4o-2024-11-20'],
    'o1': CHAT_MODEL_COST_PER_TOKEN['o1-2024-12-17'],
    'o3-mini': CHAT_MODEL_COST_PER_TOKEN['o3-mini-2025-01-31'],
}
CHAT_MODEL_COST_PER_TOKEN.update(CHAT_MODEL_COST_PER_TOKEN_PRIMARY)


EMBEDDING_MODEL_COST_PER_TOKEN = {
    # "Prices are per 1,000 tokens. You can think of tokens as pieces of words, where 1,000 tokens
    # is about 750 words. This paragraph is 35 tokens."
    # https://openai.com/pricing
    # https://platform.openai.com/docs/models
    ####
    # Embedding models
    ####
    # LATEST MODELS
    # https://openai.com/blog/new-embedding-models-and-api-updates
    'text-embedding-3-small': 0.02 / 1_000_000,
    'text-embedding-3-large': 0.13 / 1_000_000,
    # LEGACY MODELS
    'text-embedding-ada-002': 0.1 / 1_000_000,
}

MODEL_COST_PER_TOKEN = CHAT_MODEL_COST_PER_TOKEN | EMBEDDING_MODEL_COST_PER_TOKEN


@cache
def _get_encoding_for_model(model_name: str) -> Encoding:
    """Gets the encoding for a given model so that we can calculate the number of tokens."""
    return tiktoken.encoding_for_model(model_name)


def num_tokens(model_name: str, value: str) -> int:
    """For a given model, returns the number of tokens based on the str `value`."""
    return len(_get_encoding_for_model(model_name=model_name).encode(value))


def num_tokens_from_messages(model_name: str, messages: list[dict]) -> int:
    """
    Copied from https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
    Returns the number of tokens used by a list of messages.
    """
    try:
        encoding = _get_encoding_for_model(model_name)
    except KeyError:
        encoding = tiktoken.get_encoding("o200k_base")
    if model_name in {
        "gpt-3.5-turbo-0125",
        "gpt-4-0314",
        "gpt-4-32k-0314",
        "gpt-4-0613",
        "gpt-4-32k-0613",
        "gpt-4o-mini-2024-07-18",
        "gpt-4o-2024-08-06",
        }:
        tokens_per_message = 3
        tokens_per_name = 1
    elif "gpt-3.5-turbo" in model_name:
        return num_tokens_from_messages(messages=messages, model_name="gpt-3.5-turbo-0125")
    elif "gpt-4o-mini" in model_name:
        return num_tokens_from_messages(messages=messages, model_name="gpt-4o-mini-2024-07-18")
    elif "gpt-4o" in model_name:
        return num_tokens_from_messages(messages=messages, model_name="gpt-4o-2024-08-06")
    elif "gpt-4" in model_name:
        return num_tokens_from_messages(messages=messages, model_name="gpt-4-0613")
    elif "o1" in model_name or "o3-mini" in model_name:
        return num_tokens_from_messages(messages=messages, model_name="gpt-4o-2024-08-06")
    else:
        raise NotImplementedError(
            f"""num_tokens_from_messages() is not implemented for model {model_name}.""",
        )
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


def _parse_completion_chunk(chunk) -> TextChunkEvent:  # noqa: ANN001
    assert chunk.object == 'chat.completion.chunk'
    log_prob = None
    if chunk.choices[0].logprobs:
        log_prob = chunk.choices[0].logprobs.content[0].logprob
    return TextChunkEvent(
        content=chunk.choices[0].delta.content,
        logprob=log_prob,
    )


@Client.register(RegisteredClients.OPENAI)
class OpenAI(Client):
    """
    Wrapper for OpenAI API which provides a simple interface for calling the
    chat.completions.create method and parsing the response.

    The user can specify the model name, timeout, stream, and other parameters for the API call
    either in the constructor or when calling the object. If the latter, the parameters specified
    when calling the object will override the parameters specified in the constructor.
    """

    def __init__(
            self,
            model_name: str | None = None,
            reasoning_effort: ReasoningEffort | None = None,
            response_format: type[BaseModel] | None = None,
            server_url: str | None = None,
            api_key: str | None = None,
            **model_kwargs: dict,
            ) -> None:
        """
        Initialize the wrapper.

        Args:
            client:
                An instance of the AsyncOpenAI client.
            model_name:
                The model name to use for the API call (e.g. 'gpt-4o-mini').
            reasoning_effort:
                The reasoning effort to use for the API call (reasoning model required).
            response_format:
                Pydantic class defining structure of the response (i.e. structured output)
            server_url:
                The base URL for the API call.
            api_key:
                The API key to use for the API call. If not provided, the API key will be read from
                the OPENAI_API_KEY environment variable.
            **model_kwargs: Additional parameters to pass to the API call
        """
        if server_url is None and model_name not in MODEL_COST_PER_TOKEN:
            raise ValueError(f"Model '{model_name}' is not supported.")

        self.server_url = server_url
        self.client = AsyncOpenAI(
            base_url=server_url,
            api_key=api_key or os.getenv('OPENAI_API_KEY') or 'None',
        )
        self.model = model_name
        self.model_parameters = model_kwargs or {}
        self.reasoning_effort = reasoning_effort
        self.response_format = response_format
        if response_format:
            self.model_parameters['response_format'] = response_format
        if reasoning_effort:
            if isinstance(reasoning_effort, ReasoningEffort):
                self.model_parameters['reasoning_effort'] = reasoning_effort.value
            else:
                self.model_parameters['reasoning_effort'] = reasoning_effort
        # logprobs, temp, top_p are not supported with reasoning or reasoning models
        if reasoning_effort or (self.model and any(self.model.startswith(prefix) for prefix in ['o1', 'o3'])):  # noqa: E501
            self.model_parameters.pop('logprobs', None)
            self.model_parameters.pop('temperature', None)
            self.model_parameters.pop('top_p', None)

    async def stream(
            self,
            messages: list[dict],
        ) -> AsyncGenerator[TextChunkEvent | TextResponse | None]:
        """
        Streams chat chunks and returns a final summary. Note that any parameters passed to this
        method will override the parameters passed to the constructor.
        """
        model_parameters = deepcopy(self.model_parameters)
        if self.response_format:
            start_time = time.time()
            completion = await self.client.beta.chat.completions.parse(
                model=self.model,
                messages=messages,
                store=False,
                **model_parameters,
            )
            end_time = time.time()
            # these fields may not be available when using openai library for third-party providers
            if (
                hasattr(completion, 'usage')
                and hasattr(completion.usage, 'prompt_tokens_details')
                and hasattr(completion.usage.prompt_tokens_details, 'cached_tokens')
            ):
                cached_tokens = completion.usage.prompt_tokens_details.cached_tokens
                cached_cost = cached_tokens * MODEL_COST_PER_TOKEN[self.model].get('cached', 0)
            else:
                cached_tokens = None
                cached_cost = None
            yield StructuredOutputResponse(
                parsed=completion.choices[0].message.parsed,
                refusal=completion.choices[0].message.refusal,
                input_tokens=completion.usage.prompt_tokens,
                output_tokens=completion.usage.completion_tokens,
                cache_read_tokens=cached_tokens,
                input_cost=completion.usage.prompt_tokens * MODEL_COST_PER_TOKEN[self.model]['input'],  # noqa: E501
                output_cost=completion.usage.completion_tokens * MODEL_COST_PER_TOKEN[self.model]['output'],  # noqa: E501
                cache_read_cost=cached_cost,
                duration_seconds=end_time - start_time,
            )
        else:
            chunks = []
            input_tokens = 0
            output_tokens = 0
            cached_tokens = 0
            # `stream_options={'include_usage': True}` is required to get usage information
            # from API when streaming.
            # However, stream_options may not be valid when using openai library for third-party
            # providers (i.e. if server_url is None we are using the openai library)
            if (
                self.server_url is None  # if using openai library
                and (
                    # if either the user has not specified stream_options, or if they have and
                    # 'include_usage' is not set
                    'stream_options' not in model_parameters
                    or 'include_usage' not in model_parameters['stream_options']
                )
            ):
                model_parameters['stream_options'] = {'include_usage': True}
            start_time = time.time()
            response = await self.client.chat.completions.create(
                model=self.model,
                messages=messages,
                stream=True,
                store=False,
                **model_parameters,
            )
            async for chunk in response:
                if chunk.choices and chunk.choices[0].delta.content:
                    parsed_chunk = _parse_completion_chunk(chunk)
                    yield parsed_chunk
                    chunks.append(parsed_chunk)
                if hasattr(chunk, 'usage') and chunk.usage:
                    input_tokens += chunk.usage.prompt_tokens
                    output_tokens += chunk.usage.completion_tokens
                    # these fields may not be available when using openai library for third-party
                    # providers
                    if (
                        hasattr(chunk, 'usage')
                        and hasattr(chunk.usage, 'prompt_tokens_details')
                        and hasattr(chunk.usage.prompt_tokens_details, 'cached_tokens')
                    ):
                        cached_tokens += chunk.usage.prompt_tokens_details.cached_tokens

            end_time = time.time()
            if self.model in MODEL_COST_PER_TOKEN:
                input_cost = input_tokens * MODEL_COST_PER_TOKEN[self.model]['input']
                output_cost = output_tokens * MODEL_COST_PER_TOKEN[self.model]['output']
                cache_cost = cached_tokens * MODEL_COST_PER_TOKEN[self.model].get('cached', 0)
            else:
                input_cost = None
                output_cost = None
                cache_cost = None
            yield TextResponse(
                response=''.join([chunk.content for chunk in chunks]),
                input_tokens=input_tokens,
                output_tokens=output_tokens,
                cache_read_tokens=cached_tokens,
                input_cost=input_cost,
                output_cost=output_cost,
                cache_read_cost=cache_cost,
                duration_seconds=end_time - start_time,
            )


@Client.register(RegisteredClients.OPENAI_TOOLS)
class OpenAITools(Client):
    """Wrapper for OpenAI API function/tool calling."""

    def __init__(
            self,
            model_name: str,
            tools: list[Tool],
            tool_choice: ToolChoice = ToolChoice.REQUIRED,
            server_url: str | None = None,
            api_key: str | None = None,
            **model_kwargs: dict,
            ) -> None:
        """
        Initialize the wrapper.

        Args:
            model_name:
                The model name to use for the API call (e.g. 'gpt-4').
            tools:
                List of Tool objects defining available tools.
            tool_choice:
                Controls if tools are required or optional.
            server_url:
                The base URL for the API call.
            api_key:
                The API key to use for the API call. If not provided, the API key will be read from
                the OPENAI_API_KEY environment variable.
            **model_kwargs:
                Additional parameters to pass to the API call
        """
        if server_url is None and model_name not in MODEL_COST_PER_TOKEN:
            raise ValueError(f"Model '{model_name}' is not supported.")

        self.client = AsyncOpenAI(
            base_url=server_url,
            api_key=api_key or os.getenv('OPENAI_API_KEY') or 'None',
        )
        self.model = model_name
        self.model_parameters = model_kwargs or {}
        if 'temperature' not in self.model_parameters:
            self.model_parameters['temperature'] = 0.2

        if tool_choice == ToolChoice.REQUIRED:
            self.model_parameters['tool_choice'] = 'required'
        elif tool_choice == ToolChoice.AUTO:
            self.model_parameters['tool_choice'] = 'auto'
        else:
            raise ValueError(f"Invalid tool_choice: `{tool_choice}`")
        self.model_parameters['tools'] = [t.to_openai() for t in tools]

    async def stream(self, messages: list[dict[str, str]]) -> AsyncGenerator[ToolPredictionResponse | None]:  # noqa: E501
        """
        Call the model with tools.

        Args:
            messages: List of messages to send to the model.
        """
        model_parameters = deepcopy(self.model_parameters)
        start = time.time()
        completion = await self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            store=False,
            # i'm not sure it makes sense to stream chunks for tools, perhaps this will change
            # in the future; but seems overly complicated for a tool call.
            stream=False,
            **model_parameters,
        )
        end = time.time()
        # Calculate costs
        input_tokens = completion.usage.prompt_tokens
        output_tokens = completion.usage.completion_tokens
        input_cost = input_tokens * MODEL_COST_PER_TOKEN[self.model]['input']
        output_cost = output_tokens * MODEL_COST_PER_TOKEN[self.model]['output']

        # these fields may not be available when using openai library for third-party providers
        if (
            hasattr(completion, 'usage')
            and hasattr(completion.usage, 'prompt_tokens_details')
            and hasattr(completion.usage.prompt_tokens_details, 'cached_tokens')
        ):
            cached_tokens = completion.usage.prompt_tokens_details.cached_tokens
            cached_cost = cached_tokens * MODEL_COST_PER_TOKEN[self.model].get('cached', 0)
        else:
            cached_tokens = None
            cached_cost = None

        if completion.choices[0].message.tool_calls:
            message = None
            tool_call = completion.choices[0].message.tool_calls[0]
            tool_prediction = ToolPrediction(
                name=tool_call.function.name,
                arguments=json.loads(tool_call.function.arguments),
                call_id=tool_call.id,
            )
        else:
            message = completion.choices[0].message.content
            tool_prediction = None

        yield ToolPredictionResponse(
            tool_prediction=tool_prediction,
            message=message,
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            cache_read_tokens=cached_tokens,
            input_cost=input_cost,
            output_cost=output_cost,
            cache_read_cost=cached_cost,
            duration_seconds=end - start,
        )



"""Test public functions from sik_llms module."""
import os
import pytest
from sik_llms import (
    create_client,
    user_message,
    TextChunkEvent,
    TextResponse,
    Tool,
    RegisteredClients,
    ToolPredictionResponse,
    ToolPrediction,
)
from tests.conftest import ANTHROPIC_TEST_MODEL, OPENAI_TEST_MODEL


@pytest.mark.asyncio
async def test__create_client__openai() -> None:
    """Tests that model can be called from create_client."""
    client = create_client(model_name=OPENAI_TEST_MODEL)
    assert client.client is not None
    assert client.client.api_key
    assert client.client.api_key != 'None'
    assert client.client.base_url
    assert client.model == OPENAI_TEST_MODEL

    responses = []
    async for response in client.stream(messages=[user_message("What is the capital of France?")]):
        if isinstance(response, TextChunkEvent):
            responses.append(response)

    assert len(responses) > 0
    assert 'Paris' in ''.join([response.content for response in responses])

    response = client(messages=[user_message("What is the capital of France?")])
    assert isinstance(response, TextResponse)
    assert 'Paris' in response.response


@pytest.mark.skipif(os.getenv('ANTHROPIC_API_KEY') is None, reason="ANTHROPIC_API_KEY is not set")
@pytest.mark.asyncio
async def test__create_client__anthropic() -> None:
    """Tests that model can be called from create_client."""
    client = create_client(model_name=ANTHROPIC_TEST_MODEL)
    assert client.client is not None
    assert client.client.api_key
    assert client.client.api_key != 'None'
    assert client.client.base_url
    assert client.model == ANTHROPIC_TEST_MODEL

    responses = []
    async for response in client.stream(messages=[user_message("What is the capital of France?")]):
        if isinstance(response, TextChunkEvent):
            responses.append(response)

    assert len(responses) > 0
    assert 'Paris' in ''.join([response.content for response in responses])

    response = client(messages=[user_message("What is the capital of France?")])
    assert isinstance(response, TextResponse)
    assert 'Paris' in response.response


@pytest.mark.asyncio
class TestOpenAIFunctions:
    """Test the OpenAIFunctions class."""

    @pytest.mark.parametrize('is_async', [True, False])
    async def test_single_function_single_parameter__instantiate(
            self,
            simple_weather_tool: Tool,
            is_async: bool,
        ):
        """Test calling a simple function with one required parameter."""
        client = create_client(
            client_type=RegisteredClients.OPENAI_TOOLS,
            model_name=OPENAI_TEST_MODEL,
            tools=[simple_weather_tool],
        )
        if is_async:
            response = await client.run_async(
                messages=[
                    user_message("What's the weather like in Paris?"),
                ],
            )
        else:
            response = client(
                messages=[
                    user_message("What's the weather like in Paris?"),
                ],
            )
        assert isinstance(response, ToolPredictionResponse)
        assert isinstance(response.tool_prediction, ToolPrediction)
        assert response.tool_prediction.name == "get_weather"
        assert "location" in response.tool_prediction.arguments
        assert "Paris" in response.tool_prediction.arguments["location"]
        assert response.input_tokens > 0
        assert response.output_tokens > 0
        assert response.input_cost > 0
        assert response.output_cost > 0


@pytest.mark.skipif(os.getenv('ANTHROPIC_API_KEY') is None, reason="ANTHROPIC_API_KEY is not set")
@pytest.mark.asyncio
class TestAnthropicFunctions:
    """Test the AnthropicFunctions class."""

    @pytest.mark.parametrize('is_async', [True, False])
    async def test_single_function_single_parameter__instantiate(
            self,
            simple_weather_tool: Tool,
            is_async: bool,
        ):
        """Test calling a simple function with one required parameter."""
        client = create_client(
            client_type=RegisteredClients.ANTHROPIC_TOOLS,
            model_name=ANTHROPIC_TEST_MODEL,
            tools=[simple_weather_tool],
        )
        if is_async:
            response = await client.run_async(
                messages=[
                    user_message("What's the weather like in Paris?"),
                ],
            )
        else:
            response = client(
                messages=[
                    user_message("What's the weather like in Paris?"),
                ],
            )
        assert isinstance(response, ToolPredictionResponse)
        assert isinstance(response.tool_prediction, ToolPrediction)
        assert response.tool_prediction.name == "get_weather"
        assert "location" in response.tool_prediction.arguments
        assert "Paris" in response.tool_prediction.arguments["location"]
        assert response.input_tokens > 0
        assert response.output_tokens > 0
        assert response.input_cost > 0
        assert response.output_cost > 0
